import os
import numpy as np
import iris
import iris.cube
import iris.analysis
import warnings
import numba
import datetime
from numba import jit, njit
from numba import vectorize
from numba import guvectorize
from numba import cuda
from numba import int32, int64, float32, float64
from numba.cuda.cudadrv import driver

MESH_FILE = "/esarchive/autosubmit/con_files/mesh_mask_nemo.Ec3.2_O1L75.nc"
TOS_FILE = "/esarchive/exp/ecearth/a13c/cmorfiles/CMIP/EC-Earth-Consortium"\
            "/EC-Earth3-LR/historical/r1i1p1f1/Omon/tos/gn/v20180501"\
            "/tos_Omon_EC-Earth3-LR_historical_r1i1p1f1_gn_195001-195012.nc"
THETAO_FILE = "/esarchive/exp/ecearth/a13c/cmorfiles/CMIP/EC-Earth-Consortium"\
              "/EC-Earth3-LR/historical/r1i1p1f1/Omon/thetao/gn/v20180501/"\
              "thetao_Omon_EC-Earth3-LR_historical_r1i1p1f1_gn_"\
              "195001-195012.nc"
REGIONS_FILE = "/esarchive/autosubmit/con_files/mask.regions.Ec3.2_O1L75.nc"


def main():
    e1t, e2t, e3t_0 = load_mesh()
    basins = load_masks()
    tos = load_tos()
    sum_cpu = compute_regsum_2D_cpu(tos, basins, e1, e2t)
    sum_gpu = compute_regsum_2D_gpu(tos, basins, e1t, e2t)


def load_mesh():
    e1t = iris.load_cube(MESH_FILE, 'e1t').data.astype(np.float32)
    e1t = np.squeeze(e1t)
    e2t = iris.load_cube(MESH_FILE, 'e2t').data.astype(np.float32)
    e2t = np.squeeze(e2t)
    e3t = iris.load_cube(MESH_FILE, 'e3t_0').data.astype(np.float32)
    e3t = np.squeeze(e3t)
    return e1t, e2t, e3t


def load_masks():
    basins = {}
    cubes = iris.load(REGIONS_FILE)
    for cube in cubes:
        name = cube.name()
        if name in ('nav_lat', 'nav_lon', 'Caspian_Sea'):
            continue
        cube = cube.extract(iris.Constraint(z=1))
        basins[name] = cube.data.astype(np.float32)
    return basins


def load_tos():
    tos = iris.load_cube(TOS_FILE, 'sea_surface_temperature')
    tos_data = tos.data.astype(np.float32)
    return tos_data


def compute_regsum_2D_cpu(var, basins, e1t, e2t):
    regsum = {}
    for basin, mask in basins.items():
        weighted_var = compute_weighted_var_2D(var, mask, e1t, e2t)
        regsum[basin] = np.sum(weighted_var, axis=(1, 2))
    return regsum


def compute_regsum_3D_cpu(var, basins, e1t, e2t, e3t, save3d):
    regsum = {}
    regsum_3D = {}
    for basin, mask in basins.items():
        weighted_var = compute_weighted_var_3D(var, mask, e1t, e2t, e3t)
        regsum[basin] = np.sum(weighted_var, axis=(2, 3))
        if save3d:
            regsum_3D[basin] = np.sum(weighted_var, axis=(1, 2, 3))
        else:
            regsum_3D = None
    return regsum, regsum_3D


def compute_regsum_2D_gpu(var, basins, e1t, e2t):
    times = var.shape[0]
    lats = var.shape[1]
    lons = var.shape[2]
    mask = len(basins)

    masks = np.empty((mask, lats, lons))
    for i, basin in enumerate(basins):
        masks[i, :, :] = basins[basin]

    block = (128, 1, 1)
    grid_size = (lons // block[0]) + 1
    grid = (grid_size, lats, times)

    gpu_var = cuda.to_device(var.astype(np.float32))
    gpu_e1t = cuda.to_device(e1t.astype(np.float32))
    gpu_e2t = cuda.to_device(e2t.astype(np.float32))
    gpu_masks = cuda.to_device(masks)
    gpu_weighted_var = cuda.device_array((mask, times, lats*lons),
                                         dtype=np.float32)

    variable_weight_cuda[grid, block](gpu_var, gpu_e1t, gpu_e2t, gpu_masks,
                                      gpu_weighted_var, gpu_vector_weight)

    regsum = {}
    for m in range(mask):
        regmean[m] = {}
        for t in range(times):
            regmean[m][t] = sum_red_cuda(gpu_weighted_var[m, t, :]

    return regmean


@vectorize(['float32(float32, float32, float32, float32, float32)'],
           target='cpu')
def compute_weighted_var_3D(var, mask, e1t, e2t, e3t):
    weighted_var = var*mask*e1t*e2t*e3t
    return weighted_var


@vectorize(['float32(float32, float32, float32, float32)'], target='cpu')
def compute_weighted_var_2D(var, mask, e1t, e2t):
    weighted_var = var*mask*e1t*e2t
    return weighted_var


@cuda.jit
def variable_weight_cuda(var, e1t, e2t, masks, weighted_var):
    i, j, t = cuda.grid(3)
    basins = masks.shape[0]
    x = i+j*var.shape[2]
    if(i >= var.shape[2]):
        return
    for bas in range(basins):
        weighted_var[bas, t, x] = 0.0
        weighted_var[bas, t, x] = (var[t, j, i] * e1t[j, i]* e2t[j, i] *
                                   masks[bas, j, i])


if __name__ == '__main__':
    main()
